"""
Helper functions used to group interfaces together and generate the
necessary data to generate the dashboards from templates.
"""
import logging
import json
from itertools import product
from functools import reduce
from string import ascii_uppercase
from brian_dashboard_manager.templating.render import create_panel, \
    create_panel_target, create_dropdown_panel

PANEL_HEIGHT = 12
PANEL_WIDTH = 24

logger = logging.getLogger(__file__)


def num_generator(start=1):
    num = start
    while True:
        yield num
        num += 1


def gridPos_generator(id_generator, start=0, agg=False):
    num = start
    while True:
        yield {
            "height": PANEL_HEIGHT,
            "width": PANEL_WIDTH if not agg else PANEL_WIDTH // 2,
            "x": 0,
            "y": num * PANEL_HEIGHT,
            "id": next(id_generator)
        }
        if agg:
            yield {
                "height": PANEL_HEIGHT,
                "width": PANEL_WIDTH // 2,
                "x": PANEL_WIDTH // 2,
                "y": num * PANEL_HEIGHT,
                "id": next(id_generator)
            }
        num += 1


def letter_generator():
    i = 0
    j = 0
    num_letters = len(ascii_uppercase)
    while True:
        result = ascii_uppercase[i % num_letters]

        # tack on an extra letter if we are out of them
        if (i >= num_letters):
            result += ascii_uppercase[j % num_letters]
            j += 1
            if (j != 0 and j % num_letters == 0):
                i += 1
        else:
            i += 1

        yield result


def get_nren_interface_data(interfaces):
    """
    Helper for grouping interfaces into groups of NRENs
    Extracts information from interfaces to be used in panels.
    NREN dashboards have aggregate panels at the top and
    dropdowns for services / physical interfaces.
    """
    result = {}

    for interface in interfaces:

        description = interface['description'].strip()
        interface_name = interface['name']
        host = interface['router']

        router = host.replace('.geant.net', '')
        panel_title = f"{router} - {{}} - {interface_name} - {description}"

        info = interface['dashboard_info']

        dashboard_name = info['name']

        dashboard = result.get(dashboard_name, {
            'AGGREGATES': [],
            'SERVICES': [],
            'PHYSICAL': []
        })

        location = host.split('.')[1].upper()

        if info['interface_type'] == 'AGGREGATE':
            dashboard['AGGREGATES'].append({
                'interface': interface_name,
                'hostname': host,
                'alias': f"{location} - {dashboard_name} ({interface_name})"
            })

            # link aggregates are also shown
            # under the physical dropdown
            dashboard['PHYSICAL'].append({
                'title': panel_title,
                'hostname': host,
                'interface': interface_name
            })

        elif info['interface_type'] == 'LOGICAL':
            dashboard['SERVICES'].append({
                'title': panel_title,
                'hostname': host,
                'interface': interface_name
            })
        elif info['interface_type'] == 'PHYSICAL':
            dashboard['PHYSICAL'].append({
                'title': panel_title,
                'hostname': host,
                'interface': interface_name
            })

        result[dashboard_name] = dashboard
    return result


def get_interface_data(interfaces):
    """
    Helper for grouping interfaces into dashboards.
    Extracts information from interfaces to be used in panels.
    """
    result = {}

    for interface in interfaces:

        description = interface['description'].strip()
        interface_name = interface['name']
        host = interface['router']

        router = host.replace('.geant.net', '')
        panel_title = f"{router} - {{}} - {interface_name} - {description}"

        info = interface['dashboard_info']

        dashboard_name = info['name']

        dashboard = result.get(dashboard_name, [])
        dashboard.append({
            'title': panel_title,
            'interface': interface_name,
            'hostname': host,
            'has_v6': len(interface.get('ipv6', [])) > 0
        })

        result[dashboard_name] = dashboard
    return result


def get_aggregate_interface_data(interfaces, agg_type):
    """
    Helper for grouping interfaces into groups of remotes
    (ISP/NREN/...) used for aggregate dashboards
    Extracts information from interfaces to be used in panels.

    Aggregate dashboards have aggregates at the top for all remotes
    as well as aggregate panels for specific remotes.
    This builds a dict with interfaces for each remote
    and one with all interfaces.
    """

    result = []

    def reduce_func(prev, curr):
        remotes = prev.get(curr['remote'], [])
        remotes.append(curr)
        all_agg = prev.get('EVERYSINGLETARGET', [])
        all_agg.append(curr)
        prev[curr['remote']] = remotes
        prev['EVERYSINGLETARGET'] = all_agg
        return prev

    for interface in interfaces:

        interface_name = interface.get('name')
        host = interface.get('router', '')

        remote = interface['dashboard_info']['name']

        location = host.split('.')[1].upper()
        result.append({
            'type': agg_type,
            'interface': interface_name,
            'hostname': host,
            'remote': remote,
            'alias': f"{location} - {remote} ({interface_name})",
        })
    return reduce(reduce_func, result, {})


def get_aggregate_targets(targets):
    """
    Helper used for generating panel fields for aggregate panels
    with multiple target fields (ingress/egress)
    """
    ingress = []
    egress = []

    # used to generate refIds
    letters = letter_generator()

    for target in targets:
        ref_id = next(letters)
        in_data = {
            **target,
            'alias': f"{target['alias']} - Ingress Traffic",
            'refId': ref_id,
            'select_field': 'ingress'
        }
        out_data = {
            **target,
            'alias': f"{target['alias']} - Egress Traffic",
            'refId': ref_id,
            'select_field': 'egress'
        }
        ingress_target = create_panel_target(in_data)
        egress_target = create_panel_target(out_data)
        ingress.append(ingress_target)
        egress.append(egress_target)

    return ingress, egress


def get_panel_fields(panel, panel_type, datasource):
    """
    Helper for generating a single panel,
    with ingress/egress and percentile targets
    """
    letters = letter_generator()

    def get_target_data(alias, field):
        return {
            # panel includes identifying information
            # such as hostname, interface, etc.
            **panel,
            'alias': alias,
            'refId': next(letters),
            'select_field': field,
            'percentile': 'percentile' in alias.lower(),
        }

    error_fields = [('Ingress Errors', 'errorsIn'),
                    ('Egress Errors', 'errorsOut'),
                    ('Ingress Discards', 'discardsIn'),
                    ('Egress Discards', 'discardsOut')]

    ingress = ['Ingress Traffic', 'Ingress 95th Percentile']
    egress = ['Egress Traffic', 'Egress 95th Percentile']

    is_v6 = panel_type == 'IPv6'
    is_error = panel_type == 'errors'
    in_field = 'ingressv6' if is_v6 else 'ingress'

    out_field = 'egressv6' if is_v6 else 'egress'

    fields = [*product(ingress, [in_field]), *product(egress, [out_field])]

    targets = error_fields if is_error else fields

    return create_panel({
        **panel,
        'datasource': datasource,
        'linewidth': 1,
        'title': panel['title'].format(panel_type),
        'panel_targets': [get_target_data(*target) for target in targets],
        'y_axis_type': 'errors' if is_error else 'bits',
    })


def default_interface_panel_generator(gridPos):
    """
    Shared wrapper for shorter calls without
    gridPos to generate panels.

    Generates panels used in a normal dashboard
    for all traffic + (conditionally) IPv6 + Errors
    """
    def get_panel_definitions(panels, datasource, errors=False):
        result = []

        for panel in panels:
            result.append(get_panel_fields({
                **panel,
                **next(gridPos)
            }, 'traffic', datasource))
            if panel.get('has_v6', False):
                result.append(get_panel_fields({
                    **panel,
                    **next(gridPos)
                }, 'IPv6', datasource))
            if errors:
                result.append(get_panel_fields({
                    **panel,
                    **next(gridPos)
                }, 'errors', datasource))

        return result

    return get_panel_definitions


def get_nren_dashboard_data(data, datasource, tag):
    """
    Generates all panels used in a NREN dashboard,
    including dropdowns and aggregate panels.
    """

    for nren, dash in data.items():
        id_gen = num_generator()

        gridPos = gridPos_generator(id_gen, start=1)

        panel_gen = default_interface_panel_generator(gridPos)

        if len(dash['AGGREGATES']) > 0:
            agg_panels = create_aggregate_panel(
                f'Aggregate - {nren}',
                gridPos_generator(id_gen, agg=True),
                dash['AGGREGATES'], datasource)
        else:
            # if there's no aggregate panel(s), start other stuff at y=0.
            gridPos = gridPos_generator(id_gen, start=0)
            agg_panels = []

        services_dropdown = create_dropdown_panel('Services', **next(gridPos))
        service_panels = panel_gen(dash['SERVICES'], datasource)
        iface_dropdown = create_dropdown_panel('Interfaces', **next(gridPos))
        phys_panels = panel_gen(dash['PHYSICAL'], datasource, True)

        result = {
            'nren_name': nren,
            'datasource': datasource,
            'aggregate_panels': agg_panels,
            'dropdown_groups': [
                {
                    'dropdown': services_dropdown,
                    'panels': service_panels,
                },
                {
                    'dropdown': iface_dropdown,
                    'panels': phys_panels,
                }
            ]
        }
        if isinstance(tag, list):
            result['tags'] = tag
        else:
            result['tag'] = tag

        yield result


def get_dashboard_data(
        data, datasource, tag,
        panel_generator=default_interface_panel_generator,
        errors=False):
    """
    Generates all panels used in a normal dashboard without aggregate panels
    """
    id_gen = num_generator()
    gridPos = gridPos_generator(id_gen)
    panel_gen = panel_generator(gridPos)

    for dashboard_name, panels in data.items():
        result = {
            'title': dashboard_name,
            'datasource': datasource,
            'panels': list(panel_gen(panels, datasource, errors)),
        }

        if isinstance(tag, list):
            result['tags'] = tag
        else:
            result['tag'] = tag

        yield result


def create_aggregate_panel(title, gridpos, targets, datasource):
    """
    Generates a single panel with multiple targets.
    Each target is one interface / line on the graph
    """

    ingress_targets, egress_targets = get_aggregate_targets(targets)
    result = []

    ingress_pos = next(gridpos)
    egress_pos = next(gridpos)

    is_total = 'totals' in title.lower()

    def reduce_alias(prev, curr):
        d = json.loads(curr)
        alias = d['alias']
        if 'egress' in alias.lower():
            prev[alias] = '#0000FF'
        else:
            prev[alias] = '#00FF00'
        return prev

    ingress_colors = reduce(reduce_alias, ingress_targets, {})
    egress_colors = reduce(reduce_alias, egress_targets, {})

    result.append(create_panel({
        **ingress_pos,
        'stack': True,
        'linewidth': 0 if is_total else 1,
        'datasource': datasource,
        'title': title + ' - ingress',
        'targets': ingress_targets,
        'y_axis_type': 'bits',
        'alias_colors': json.dumps(ingress_colors) if is_total else {}
    }))

    result.append(create_panel({
        **egress_pos,
        'stack': True,
        'linewidth': 0 if is_total else 1,
        'datasource': datasource,
        'title': title + ' - egress',
        'targets': egress_targets,
        'y_axis_type': 'bits',
        'alias_colors': json.dumps(egress_colors) if is_total else {}
    }))

    return result


def get_aggregate_dashboard_data(title, targets, datasource, tag):
    """
    Creates three types of aggregate panels:
    Aggregate Ingress/Egress that contain
    every target (interface) given as parameter

    Totals Ingress/Egress which is the same as above,
    but with a different line color.

    Aggregates for each remote
    (all interfaces for each remote (ISP/NREN/...)
    on separate graphs
    """

    id_gen = num_generator()
    gridPos = gridPos_generator(id_gen, agg=True)

    panels = []
    all_targets = targets.get('EVERYSINGLETARGET', [])

    ingress, egress = create_aggregate_panel(
        title, gridPos, all_targets, datasource)
    panels.extend([ingress, egress])

    totals_title = title + ' - Totals'
    t_in, t_eg = create_aggregate_panel(
        totals_title, gridPos, all_targets, datasource)
    panels.extend([t_in, t_eg])

    if 'EVERYSINGLETARGET' in targets:
        del targets['EVERYSINGLETARGET']

    for target in targets:
        _in, _out = create_aggregate_panel(
            title + f' - {target}', gridPos, targets[target], datasource)
        panels.extend([_in, _out])

    result = {
        'title': title,
        'datasource': datasource,
        'panels': panels,

    }
    if isinstance(tag, list):
        result['tags'] = tag
    else:
        result['tag'] = tag

    return result
